import datetime
import pytz
import boto3
import logging
from time import sleep
from autom8it import Log
from autom8it import AutomationTask

UTC = pytz.UTC

for log_name in ['boto', 'boto3', 'botocore', 's3transfer', 'urllib3']:
    logging.getLogger(log_name).setLevel(logging.WARNING)


class DeregisterTarget(AutomationTask):

    TARGET_GROUP_ARN_KEY = 'target_group_arn'
    TARGET_ID_KEY = 'target_id'
    PORT_KEY = 'port'

    aws_client = boto3.client('elbv2')

    def __init__(self, task_data: dict):
        super().__init__(task_data=task_data, check_done_interval=10)

    @property
    def validation_schema(self) -> dict:
        return {
            self.TARGET_GROUP_ARN_KEY: {
                'required': True,
                'type': 'string'
            },
            self.TARGET_ID_KEY: {
                'required': True,
                'type': 'string'
            },
            self.PORT_KEY: {
                'required': True,
                'type': 'integer'
            }
        }

    @property
    def task_type(self) -> str:
        return 'Deregister target'

    def do(self):
        target_group_desc = self.aws_client.deregister_targets(
            TargetGroupArn=self.get_task_attribute(self.TARGET_GROUP_ARN_KEY),
            Targets=[{
                'Id': self.get_task_attribute(self.TARGET_ID_KEY),
                'Port': self.get_task_attribute(self.PORT_KEY)
            }],
        )
        return target_group_desc

    def is_done(self) -> bool:
        target_id = self.get_task_attribute(self.TARGET_ID_KEY)
        target_group_desc = self.aws_client.describe_target_health(
            TargetGroupArn=self.get_task_attribute(self.TARGET_GROUP_ARN_KEY)
        )
        for target in target_group_desc['TargetHealthDescriptions']:
            if target['Target']['Id'] == target_id:
                Log.debug(f"Target {target_id} state: {target['TargetHealth']['State']}")
                return False

        return True


class RegisterTarget(DeregisterTarget):

    def __init__(self, task_data: dict):
        super().__init__(task_data=task_data)

    @property
    def task_type(self) -> str:
        return 'Register target'

    def do(self):
        target_group_desc = self.aws_client.register_targets(
            TargetGroupArn=self.get_task_attribute(self.TARGET_GROUP_ARN_KEY),
            Targets=[{
                'Id': self.get_task_attribute(self.TARGET_ID_KEY),
                'Port': self.get_task_attribute(self.PORT_KEY, 80)
            }],
        )
        return target_group_desc

    def is_done(self) -> bool:
        target_id = self.get_task_attribute(self.TARGET_ID_KEY)
        target_group_desc = self.aws_client.describe_target_health(
            TargetGroupArn=self.get_task_attribute(self.TARGET_GROUP_ARN_KEY)
        )
        for target in target_group_desc['TargetHealthDescriptions']:
            if target['Target']['Id'] == target_id:
                target_state = target['TargetHealth']['State']
                Log.debug(f"Target {target_id} state: {target_state}")
                if target_state in ['healthy']:
                    return True

        return False


class StopEC2Instance(AutomationTask):

    INSTANCE_ID_KEY = 'instance_id'

    @property
    def validation_schema(self) -> dict:
        return {
            self.INSTANCE_ID_KEY: {
                'required': True,
                'type': 'string'
            },
        }

    aws_client = boto3.client('ec2')

    def __init__(self, task_data: dict):
        super().__init__(task_data=task_data)

    @property
    def task_type(self) -> str:
        return 'Stop EC2 instance'

    def do(self):
        return self.aws_client.stop_instances(
            InstanceIds=[self.get_task_attribute(self.INSTANCE_ID_KEY)],
        )

    def is_done(self) -> bool:
        return self.is_state_equals(requested_state='stopped')

    def is_state_equals(self, requested_state: str) -> bool:
        instance_id = self.get_task_attribute(self.INSTANCE_ID_KEY)
        desc = self.aws_client.describe_instance_status(
            InstanceIds=[instance_id],
        )
        for data in desc['InstanceStatuses']:
            if data['InstanceId'] == instance_id:
                state = data['InstanceState']['Name']
                Log.debug(f"Instance {instance_id} state: {state}")
                if state in [requested_state]:
                    return True

        return False


class StartEC2Instance(StopEC2Instance):

    def __init__(self, task_data: dict):
        super().__init__(task_data=task_data)

    @property
    def task_type(self) -> str:
        return 'Start EC2 instance'

    def do(self):
        return self.aws_client.start_instances(
            InstanceIds=[self.get_task_attribute(self.INSTANCE_ID_KEY)],
        )

    def is_done(self) -> bool:
        return self.is_state_equals(requested_state='running')


class RebootEC2Instance(StopEC2Instance):

    def __init__(self, task_data: dict):
        super().__init__(task_data=task_data)

    @property
    def task_type(self) -> str:
        return 'Start EC2 instance'

    def do(self):
        resp = self.aws_client.reboot_instances(
            InstanceIds=[self.get_task_attribute(self.INSTANCE_ID_KEY)],
        )
        sleep(90.0)
        return resp

    def is_done(self) -> bool:
        return self.is_state_equals(requested_state='running')


class VerifyECSServicesCount(AutomationTask):

    CLUSTER_KEY = 'cluster'
    SERVICES_KEY = 'services'
    SERVICE_RUNNING_COUNT_KEY = 'runningCount'
    SERVICE_DESIRED_COUNT_KEY = 'desiredCount'

    @property
    def validation_schema(self) -> dict:
        return {
            self.CLUSTER_KEY: {
                'required': True,
                'type': 'string'
            },
            self.SERVICES_KEY: {
                'required': True,
                'type': 'list'
            },
        }

    ecs_client = boto3.client('ecs')

    def __init__(self, task_data: dict):
        super().__init__(task_data=task_data)

    @property
    def task_type(self) -> str:
        return 'Verify ECS services count'

    def describe_services(self) -> dict:
        return self.ecs_client.describe_services(
            cluster=self.get_task_attribute(self.CLUSTER_KEY),
            services=self.get_task_attribute(self.SERVICES_KEY),
        )

    def do(self):
        return 'OK'

    def is_done(self) -> bool:
        desc = self.describe_services()
        for service in desc[self.SERVICES_KEY]:
            desired_count = service[self.SERVICE_DESIRED_COUNT_KEY]
            running_count = service[self.SERVICE_RUNNING_COUNT_KEY]
            if desired_count != running_count:
                Log.debug(
                    f"Service {service['serviceName']} is not ready, "
                    f"desired count is {desired_count} but running count is {running_count}."
                )
                return False
        return True


class UpdateServicesWithLastRevision(VerifyECSServicesCount):

    def __init__(self, task_data: dict):
        super().__init__(task_data=task_data)

    @property
    def task_type(self) -> str:
        return 'Update ECS services with last revision'

    def do(self):
        result = []
        for service in self.get_task_attribute(self.SERVICES_KEY):

            desc = self.ecs_client.update_service(
                cluster=self.get_task_attribute(self.CLUSTER_KEY),
                service=service,
                taskDefinition=self._get_task_definition_family(service=service)
            )
            result.append(desc)

        return result

    def _get_task_definition_family(self, service: str) -> str:
        desc = self.ecs_client.describe_services(
            cluster=self.get_task_attribute(self.CLUSTER_KEY),
            services=[service]
        )
        services_desc = desc['services']
        Log.debug(f'Services description: {services_desc}')
        if len(services_desc) == 1:
            return services_desc[0]['taskDefinition'].split('/')[1].split(':')[0]
        else:
            raise ValueError(f'Expected only one service, got {len(services_desc)}.')

    def is_done(self) -> bool:
        return True


class ChangeTaskDefinitionContainerImage(AutomationTask):

    TASK_DEFINITION_KEY = 'task_definition'
    IMAGE_KEY = 'image'
    CONTAINER_DEFINITIONS_KEY = 'containerDefinitions'

    @property
    def validation_schema(self) -> dict:
        return {
            self.TASK_DEFINITION_KEY: {
                'required': True,
                'type': 'string'
            },
            self.IMAGE_KEY: {
                'type': 'string'
            }
        }

    aws_ecs_client = boto3.client('ecs')
    aws_ecr_client = boto3.client('ecr')

    def __init__(self, task_data: dict):
        super().__init__(task_data=task_data)

    @property
    def task_type(self) -> str:
        return 'Create ECS Task Definition revision'

    def do(self):
        desc = self.aws_ecs_client.describe_task_definition(
            taskDefinition=self.get_task_attribute(self.TASK_DEFINITION_KEY)
        )
        Log.debug(f'Task definition description: {desc}')
        container_definitions = desc['taskDefinition'][self.CONTAINER_DEFINITIONS_KEY]
        for container_def in container_definitions:
            image_uri = self.get_task_attribute(self.IMAGE_KEY, default=None)
            if image_uri is None:
                image_uri = self.get_last_image_tag(
                    repository=self.get_repository_name(image_uri=container_def[self.IMAGE_KEY])
                )
            if ':' not in image_uri:
                image_uri = f"{container_def[self.IMAGE_KEY].split(':')[0]}:{image_uri}"
            container_def[self.IMAGE_KEY] = image_uri

        task_definition_desc = desc.get('taskDefinition', {})

        result = self.aws_ecs_client.register_task_definition(
            family=self.get_task_attribute(self.TASK_DEFINITION_KEY),
            containerDefinitions=container_definitions,
            taskRoleArn=task_definition_desc.get('taskRoleArn', ''),
            executionRoleArn=task_definition_desc.get('executionRoleArn', ''),
            networkMode=task_definition_desc.get('networkMode'),
            volumes=task_definition_desc.get('volumes', []),
            placementConstraints=task_definition_desc.get('placementConstraints', []),
            requiresCompatibilities=task_definition_desc.get('requiresCompatibilities', [])
        )

        return result

    def is_done(self) -> bool:
        return True

    @staticmethod
    def get_repository_name(image_uri: str) -> str:
        """
        Example: 112358132134.dkr.ecr.us-east-2.amazonaws.com/my-repo:1.10.21-tag
        Will return my-repo
        :param image_uri:
        :return:
        """
        return image_uri.split('/')[1].split(':')[0]

    @staticmethod
    def get_last_image_tag(repository: str) -> str:
        result = ChangeTaskDefinitionContainerImage.aws_ecr_client.describe_images(
            repositoryName=repository
        )
        images = {image['imageTags'][0]: image['imagePushedAt'] for image in result['imageDetails']}
        last_date = datetime.datetime(year=1970, month=1, day=1).replace(tzinfo=UTC)
        last_tag = None
        for tag, date in images.items():
            date = date.replace(tzinfo=UTC)
            if date > last_date:
                last_date = date
                last_tag = tag
        return last_tag


class StopServicesAndWaitUntilStarts(VerifyECSServicesCount):

    def __init__(self, task_data: dict):
        super().__init__(task_data=task_data)

    @property
    def task_type(self) -> str:
        return 'Stop all task of ECS services wait until they starts again'

    def stop_tasks(self, tasks_arn: list):
        for task_arn in tasks_arn:
            self.ecs_client.stop_task(
                cluster=self.get_task_attribute(self.CLUSTER_KEY),
                task=task_arn,
                reason=f'Stopping service by Autom8it ({self.__class__.__name__})'
            )

    def do(self):
        result = []
        for service in self.get_task_attribute(self.SERVICES_KEY):

            desc = self.ecs_client.list_tasks(
                cluster=self.get_task_attribute(self.CLUSTER_KEY),
                serviceName=service,
                desiredStatus='RUNNING',
            )
            tasks_arn = desc.get('taskArns', [])
            self.stop_tasks(tasks_arn=tasks_arn)
            result.append(tasks_arn)

        return result



import os
import json
import uuid
import inspect
import logging.handlers
from datetime import datetime

PROCESS_LOG_ID = f'{uuid.uuid4()}'

_root = logging.getLogger()

_root.setLevel(0)

default_log_file_name = f"{datetime.now().replace(microsecond=0)}_{PROCESS_LOG_ID}".replace(' ', '_').replace(':', '-')
FORMAT = '%(asctime)s > %(levelname)s\t%(message)s'
formatter = logging.Formatter(FORMAT)

QUICKBE_LOG_TO_STDOUT = os.getenv('QUICKBE_LOG_TO_STDOUT', 'True')
if QUICKBE_LOG_TO_STDOUT.upper() in ['TRUE', '1', 'YES']:
    stdout_handler = logging.StreamHandler()
    stdout_handler.setFormatter(formatter)
    _root.addHandler(stdout_handler)

LOG_MESSAGE_FORMAT = '{0}({1}) method: {2} \t {3}'


def run_level_source_file(run_level: int = 2) -> str:
    """
    Get the source file name of the calling method.
    :param run_level: Each method in the stack is adding one run level.
    :return: The name of the file that contains the method that called the LogHandler.
    """
    filename = inspect.stack()[run_level][1]
    first, *middle, last = '/prefix{0}'.format(filename).replace('\\', '/').split('/')
    return last


def run_level_source_line_number(run_level: int = 2) -> str:
    """
    Get the line number in the source file name of the calling method.
    :param run_level: Each method in the stack is adding one run level.
    :return: Line number that called the LogHandler.
    """
    return inspect.stack()[run_level][2]


def run_level_method_name(run_level: int = 2) -> str:
    """
    Get the calling method name.
    :param run_level: Each method in the stack is adding one run level.
    :return: Method name that called the LogHandler
    """
    return inspect.stack()[run_level][3]


def _object_dump(obj: object) -> str:
    return json.dumps(obj, default=lambda o: o.__dict__, indent=4)


def log_exception(message: str):
    msg = LOG_MESSAGE_FORMAT.format(
        run_level_source_file(run_level=3),
        run_level_source_line_number(run_level=3),
        run_level_method_name(run_level=3), message
    )
    logging.exception(msg=msg)


def log_msg(level: int, message: object, current_run_level: int = 2) -> None:
    """
    Write a log message.
    :param current_run_level:
    :param level: Logging level 0=NOTSET, 10=DEBUG, 20=INFO, 30=WARNING, 40=ERROR, 50=CRITICAL
    :param message: Object or string
    :return: None
    """
    if not isinstance(message, str):
        message = _object_dump(message)

    logging.log(
        level,
        LOG_MESSAGE_FORMAT.format(
            run_level_source_file(run_level=current_run_level),
            run_level_source_line_number(run_level=current_run_level),
            run_level_method_name(run_level=current_run_level), message
        )
    )


def set_log_level(level: int):
    _root.setLevel(level=level)


def get_log_level() -> int:
    return _root.level


def get_log_level_name() -> str:
    return logging.getLevelName(get_log_level())



import abc
import yaml
import uuid
import requests
from time import sleep
from datetime import datetime
from cerberus import Validator
import autom8it.logger as b_logger
from importlib import import_module


class Log:

    _stopwatches = {}
    _warning_msgs_count = 0
    _error_msgs_count = 0
    _critical_msgs_count = 0

    @staticmethod
    def set_log_level(level: int):
        b_logger.set_log_level(level=level)

    @staticmethod
    def get_log_level() -> int:
        return b_logger.get_log_level()

    @staticmethod
    def get_log_level_name() -> str:
        return b_logger.get_log_level_name()

    @staticmethod
    def debug(msg: str):
        b_logger.log_msg(level=10, message=msg, current_run_level=3)

    @staticmethod
    def info(msg: str):
        b_logger.log_msg(level=20, message=msg, current_run_level=3)

    @staticmethod
    def warning(msg: str):
        Log._warning_msgs_count += 1
        b_logger.log_msg(level=30, message=msg, current_run_level=3)

    @staticmethod
    def error(msg: str):
        Log._error_msgs_count += 1
        b_logger.log_msg(level=40, message=msg, current_run_level=3)

    @staticmethod
    def critical(msg: str):
        Log._critical_msgs_count += 1
        b_logger.log_msg(level=50, message=msg, current_run_level=3)

    @staticmethod
    def exception(msg: str):
        b_logger.log_exception(message=msg)
        # Log._critical_msgs_count += 1
        # b_logger.log_msg(level=50, message=msg, current_run_level=3)

    @staticmethod
    def warning_count() -> int:
        return Log._warning_msgs_count

    @staticmethod
    def error_count() -> int:
        return Log._error_msgs_count

    @staticmethod
    def critical_count() -> int:
        return Log._critical_msgs_count

    @staticmethod
    def start_stopwatch(msg: str, print_it: bool = False) -> str:
        stopwatch_id = str(uuid.uuid4())
        Log._stopwatches[stopwatch_id] = [datetime.now(), msg]
        if print_it:
            b_logger.log_msg(
                level=10,
                message=f'Start stopwatch: {msg}\t id={stopwatch_id}',
                current_run_level=3
            )
        return stopwatch_id

    @staticmethod
    def stopwatch_seconds(stopwatch_id: str, print_it: bool = True) -> float:
        if stopwatch_id in Log._stopwatches:
            start_time, msg = Log._stopwatches[stopwatch_id]
            time_delta = datetime.now() - start_time
            seconds = time_delta.total_seconds()
            if print_it:
                b_logger.log_msg(
                    level=10,
                    message=f'{seconds} seconds from start, {Log._stopwatches[stopwatch_id][1]}.',
                    current_run_level=3
                )
            return seconds
        else:
            return -1

    @staticmethod
    def stop_stopwatch(stopwatch_id: str, print_it: bool = False) -> bool:
        if stopwatch_id in Log._stopwatches:
            start_time, msg = Log._stopwatches[stopwatch_id]
            if print_it:
                seconds = Log.stopwatch_seconds(stopwatch_id=stopwatch_id, print_it=False)
                b_logger.log_msg(
                    level=10,
                    message=f'{msg} took {seconds} seconds.',
                    current_run_level=3
                )
            try:
                del Log._stopwatches[stopwatch_id]
            except KeyError:
                pass
            return True
        else:
            return False


class AutomationTask:

    def __init__(self, task_data: dict, timeout: int = 900, check_done_interval: int = 5):
        validator = Validator(self.validation_schema)
        validator.allow_unknown = True
        if validator.validate(task_data):
            self._task_data = task_data
            self._check_done_interval = check_done_interval
            self._timeout = timeout
        else:
            for error in validator.errors:
                Log.error(error)
            raise ValueError(f'There are {len(validator.errors)} in task data.')

    @property
    @abc.abstractmethod
    def task_type(self) -> str:
        pass

    """
    https://docs.python-cerberus.org/en/stable/validation-rules.html#
    Validation schema example
        {
            'name': {
                'required': True, 
                'type': 'string'
            }, 
            'age': {
                'type': 'integer'
                'min': 18,
                'max': 120
            }
            'gender': {
                'type': 'string',
                'allowed': ['male', 'female', 'unknown']
        }
    """
    @property
    @abc.abstractmethod
    def validation_schema(self) -> dict:
        pass

    @abc.abstractmethod
    def do(self):
        pass

    @abc.abstractmethod
    def is_done(self) -> bool:
        pass

    def get_task_attribute(self, key: str, default=None):
        return self._task_data.get(key, default)

    def run(self):
        sw_id = Log.start_stopwatch(msg=self.task_type)
        Log.debug(f"Task returned: {self.do()}")
        while not self.is_done() and self._timeout > Log.stopwatch_seconds(stopwatch_id=sw_id, print_it=False):
            sleep(self._check_done_interval)


def run_yaml(file_path: str):
    with open(file_path, "r") as f:
        run_def = yaml.safe_load(f)
    run(run_def=run_def)


TASK_CLASS_KEY = 'class'
TASK_PARAMETERS_KEY = 'parameters'


def run(run_def: dict):
    main_sw_id = Log.start_stopwatch(msg='Run')
    for task_name, task_def in run_def.items():
        if TASK_CLASS_KEY in task_def:
            task = get_task_instance(class_name=task_def[TASK_CLASS_KEY], parameters=task_def[TASK_PARAMETERS_KEY])
            first_row = f'*   Task {task_name} ({task.task_type}) is starting *'
            Log.info(msg="*" * len(first_row))
            Log.info(msg=first_row)
            Log.info(msg="*" * len(first_row))
            Log.info(f'Task parameters: {task_def[TASK_PARAMETERS_KEY]}')
            task_sw_id = Log.start_stopwatch(msg=first_row)
            task.run()
            Log.info(
                f'*** Task {task.task_type} finished after '
                f'{Log.stopwatch_seconds(stopwatch_id=task_sw_id, print_it=False)} seconds ***'
            )
        else:
            raise SyntaxError(f'Missing {TASK_CLASS_KEY} definition.')
    Log.info(f'Process finished after {Log.stopwatch_seconds(stopwatch_id=main_sw_id, print_it=False)} seconds.')


def get_task_instance(class_name: str, parameters: dict) -> AutomationTask:
    task_cls = get_class_by_name(class_name=class_name, requested_type=AutomationTask)
    return task_cls(parameters)


def get_class_by_name(class_name: str, requested_type: object = None):

    try:
        module_path, c_name = class_name.rsplit('.', 1)
        module = import_module(module_path)
        result = getattr(module, c_name)
        if requested_type is not None and not issubclass(result, requested_type):
            raise ValueError(f'Class {result} does not match type {requested_type}.')
        return result
    except (ImportError, AttributeError):
        raise ImportError(class_name)


class HttpRequestTask(AutomationTask):

    HTTP_METHOD = 'method'
    URL = 'url'
    DATA = 'data'
    HEADERS = 'headers'
    PARAMETERS = 'parameters'

    def __init__(self, task_data: dict):
        super().__init__(task_data=task_data, check_done_interval=0)

    @property
    def validation_schema(self) -> dict:
        return {
            self.HTTP_METHOD: {
                'required': True,
                'type': 'string',
                'allowed': ['post', 'POST', 'get', 'GET']
            },
            self.URL: {
                'required': True,
                'type': 'string'
            },
            self.DATA: {
                'type': ['string', 'dict']
            },
            self.HEADERS: {
                'type': 'dict'
            },
            self.PARAMETERS: {
                'type': 'dict'
            }
        }

    @property
    def task_type(self) -> str:
        return 'Send HTTP reqeust'

    def do(self):
        resp = 'Error'
        url = self.get_task_attribute(key=self.URL)
        http_method = self.get_task_attribute(key=self.HTTP_METHOD, default='get').lower()
        headers = self.get_task_attribute(key=self.HEADERS, default={})
        parameters = self.get_task_attribute(key=self.PARAMETERS, default={})
        data = self.get_task_attribute(key=self.DATA, default='')

        if http_method == 'get':
            resp = requests.get(
                url,
                headers=headers,
                params=parameters
            )
        elif http_method == 'post':
            if isinstance(data, dict):
                resp = requests.post(
                    url,
                    json=data,
                    headers=headers,
                    params=parameters
                )
            else:
                resp = requests.post(
                    url,
                    data=data,
                    headers=headers,
                    params=parameters
                )

        Log.debug(f'Task got HTTP response: {resp}, {resp.text}')
        return resp

    def is_done(self) -> bool:
        return True
